"use strict";
/**
 * The commands for the quint CLI
 *
 * See the description at:
 * https://github.com/informalsystems/quint/blob/main/doc/quint.md
 *
 * @author Igor Konnov, Gabriela Moreira, Shon Feder, Informal Systems, 2021-2023
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.outputResult = exports.docs = exports.verifySpec = exports.outputCompilationTarget = exports.compile = exports.runSimulator = exports.runTests = exports.runRepl = exports.typecheck = exports.mkErrorMessage = exports.parse = exports.load = void 0;
const fs_1 = require("fs");
const json_bigint_1 = __importDefault(require("json-bigint"));
const path_1 = require("path");
const process_1 = require("process");
const chalk_1 = __importDefault(require("chalk"));
const quintParserFrontend_1 = require("./parsing/quintParserFrontend");
const either_1 = require("@sweet-monads/either");
const assert_1 = __importStar(require("assert"));
const repl_1 = require("./repl");
const errorReporter_1 = require("./errorReporter");
const docs_1 = require("./docs");
const quintError_1 = require("./quintError");
const idGenerator_1 = require("./idGenerator");
const itf_1 = require("./itf");
const graphics_1 = require("./graphics");
const verbosity_1 = require("./verbosity");
const rng_1 = require("./rng");
const sourceResolver_1 = require("./parsing/sourceResolver");
const quintVerifier_1 = require("./quintVerifier");
const fullFlattener_1 = require("./flattening/fullFlattener");
const quintAnalyzer_1 = require("./quintAnalyzer");
const trace_1 = require("./runtime/trace");
const lodash_1 = require("lodash");
const maybe_1 = require("@sweet-monads/maybe");
const compileToTlaplus_1 = require("./compileToTlaplus");
const evaluator_1 = require("./runtime/impl/evaluator");
const IRVisitor_1 = require("./ir/IRVisitor");
// Extract just the parts of a ProcedureStage that we use for the output
// See https://stackoverflow.com/a/39333479/1187277
const pickOutputStage = ({ stage, warnings, modules, table, types, effects, errors, documentation, passed, failed, ignored, status, trace, seed, }) => {
    return {
        stage,
        warnings,
        modules,
        table,
        types,
        effects,
        errors,
        documentation,
        passed,
        failed,
        ignored,
        status,
        trace,
        seed,
    };
};
function cliErr(msg, stage) {
    return (0, either_1.left)({ msg, stage });
}
/** Load a file into a string
 *
 * @param args the CLI arguments parsed by yargs */
async function load(args) {
    const stage = { stage: 'loading', args };
    if ((0, fs_1.existsSync)(args.input)) {
        try {
            const path = (0, path_1.resolve)((0, process_1.cwd)(), args.input);
            const sourceCode = (0, fs_1.readFileSync)(path, 'utf8');
            return (0, either_1.right)({
                ...stage,
                args,
                path,
                sourceCode: new Map([[path, sourceCode]]),
                warnings: [],
            });
        }
        catch (err) {
            return cliErr(`file ${args.input} could not be opened due to ${err}`, {
                ...stage,
                errors: [],
                sourceCode: new Map(),
            });
        }
    }
    else {
        return cliErr(`file ${args.input} does not exist`, { ...stage, errors: [], sourceCode: new Map() });
    }
}
exports.load = load;
/**
 * Parse a Quint specification.
 *
 * @param loaded the procedure stage produced by `load`
 */
async function parse(loaded) {
    const { args, sourceCode, path } = loaded;
    const text = sourceCode.get(path);
    const parsing = { ...loaded, stage: 'parsing' };
    const idGen = (0, idGenerator_1.newIdGenerator)();
    return (0, lodash_1.flow)([
        () => {
            const phase1Data = (0, quintParserFrontend_1.parsePhase1fromText)(idGen, text, path);
            // if there is exactly one module in the original text, make it the main one
            const defaultModuleName = phase1Data.modules.length === 1 ? (0, maybe_1.just)(phase1Data.modules[0].name) : (0, maybe_1.none)();
            return { ...phase1Data, defaultModuleName };
        },
        phase1Data => {
            const resolver = (0, sourceResolver_1.fileSourceResolver)(sourceCode);
            const mainPath = resolver.lookupPath((0, path_1.dirname)(path), (0, path_1.basename)(path));
            return (0, quintParserFrontend_1.parsePhase2sourceResolution)(idGen, resolver, mainPath, phase1Data);
        },
        phase2Data => {
            if (args.sourceMap) {
                // Write source map to the specified file
                writeToJson(args.sourceMap, (0, quintParserFrontend_1.compactSourceMap)(phase2Data.sourceMap));
            }
            return (0, quintParserFrontend_1.parsePhase3importAndNameResolution)(phase2Data);
        },
        phase3Data => (0, quintParserFrontend_1.parsePhase4toposort)(phase3Data),
        phase4Data => ({ ...parsing, ...phase4Data, idGen }),
        result => {
            if (result.errors.length > 0) {
                const newErrorMessages = result.errors.map(mkErrorMessage(result.sourceMap));
                const errorMessages = parsing.errors ? parsing.errors.concat(newErrorMessages) : newErrorMessages;
                return (0, either_1.left)({ msg: 'parsing failed', stage: { ...result, errors: errorMessages } });
            }
            return (0, either_1.right)(result);
        },
    ])();
}
exports.parse = parse;
function mkErrorMessage(sourceMap) {
    return error => {
        const loc = error.reference ? sourceMap.get(error.reference) : undefined;
        return {
            explanation: (0, quintError_1.quintErrorToString)(error),
            locs: loc ? [loc] : [],
        };
    };
}
exports.mkErrorMessage = mkErrorMessage;
/**
 * Check types and effects of a Quint specification.
 *
 * @param parsed the procedure stage produced by `parse`
 */
async function typecheck(parsed) {
    const { table, modules, sourceMap } = parsed;
    const [errorMap, result] = (0, quintAnalyzer_1.analyzeModules)(table, modules);
    const typechecking = { ...parsed, ...result, stage: 'typechecking' };
    if (errorMap.length === 0) {
        return (0, either_1.right)(typechecking);
    }
    else {
        const errors = errorMap.map(mkErrorMessage(sourceMap));
        return cliErr('typechecking failed', { ...typechecking, errors });
    }
}
exports.typecheck = typecheck;
/**
 * Run REPL.
 *
 * @param _argv parameters as provided by yargs
 */
async function runRepl(_argv) {
    let filename = undefined;
    let moduleName = undefined;
    if (_argv.require) {
        // quint -r FILE.qnt or quint -r FILE.qnt::MODULE
        const m = /^(.*?)(?:|::([a-zA-Z_]\w*))$/.exec(_argv.require);
        if (m) {
            ;
            [filename, moduleName] = m.slice(1, 3);
        }
    }
    const options = {
        preloadFilename: filename,
        importModule: moduleName,
        verbosity: _argv.quiet ? 0 : _argv.verbosity,
    };
    (0, repl_1.quintRepl)(process.stdin, process.stdout, options);
}
exports.runRepl = runRepl;
/**
 * Run the tests. We imitate the output of mocha.
 *
 * @param typedStage the procedure stage produced by `typecheck`
 */
async function runTests(prev) {
    const testing = { ...prev, stage: 'testing' };
    const verbosityLevel = deriveVerbosity(prev.args);
    const mainName = guessMainModule(prev);
    const main = prev.modules.find(m => m.name === mainName);
    if (!main) {
        const error = { code: 'QNT405', message: `Main module ${mainName} not found` };
        return cliErr('Argument error', { ...testing, errors: [mkErrorMessage(prev.sourceMap)(error)] });
    }
    const rngOrError = mkRng(prev.args.seed);
    if (rngOrError.isLeft()) {
        return cliErr(rngOrError.value, { ...testing, errors: [] });
    }
    const rng = rngOrError.unwrap();
    const matchFun = (n) => isMatchingTest(testing.args.match, n);
    const maxSamples = testing.args.maxSamples;
    const options = {
        testMatch: matchFun,
        maxSamples: testing.args.maxSamples,
        rng,
        verbosity: verbosityLevel,
        onTrace: (index) => (name, status, vars, states) => {
            if (outputTemplate) {
                const filename = expandNamedOutputTemplate(outputTemplate, name, index, { autoAppend: prev.args.nTraces > 1 });
                const trace = (0, itf_1.toItf)(vars, states);
                if (trace.isRight()) {
                    const jsonObj = addItfHeader(prev.args.input, status, trace.value);
                    writeToJson(filename, jsonObj);
                }
                else {
                    console.error(`ITF conversion failed on ${name}: ${trace.value}`);
                }
            }
        },
    };
    const out = console.log;
    const outputTemplate = testing.args.outItf;
    // Start the Timer and being running the tests
    const startMs = Date.now();
    if (verbosity_1.verbosity.hasResults(verbosityLevel)) {
        out(`\n  ${mainName}`);
    }
    const testDefs = Array.from(prev.resolver.collector.definitionsByModule.get(mainName).values())
        .flat()
        .filter(d => d.kind === 'def' && options.testMatch(d.name));
    const evaluator = new evaluator_1.Evaluator(testing.table, (0, trace_1.newTraceRecorder)(verbosityLevel, rng, 1), rng);
    const results = testDefs.map((def, index) => {
        return evaluator.test(def, maxSamples, options.onTrace(index));
    });
    // We're finished running the tests
    const elapsedMs = Date.now() - startMs;
    // output the status for every test
    let nFailures = 1;
    if (verbosity_1.verbosity.hasResults(verbosityLevel)) {
        results.forEach(res => {
            if (res.status === 'passed') {
                out(`    ${chalk_1.default.green('ok')} ${res.name} passed ${res.nsamples} test(s)`);
            }
            if (res.status === 'failed') {
                const errNo = chalk_1.default.red(nFailures);
                out(`    ${errNo}) ${res.name} failed after ${res.nsamples} test(s)`);
                nFailures++;
            }
        });
    }
    const passed = results.filter(r => r.status === 'passed');
    const failed = results.filter(r => r.status === 'failed');
    const ignored = results.filter(r => r.status === 'ignored');
    const namedErrors = failed.reduce((acc, failure) => acc.concat(failure.errors.map(e => [failure, mkErrorMessage(testing.sourceMap)(e)])), []);
    // output the statistics banner
    if (verbosity_1.verbosity.hasResults(verbosityLevel)) {
        out('');
        if (passed.length > 0) {
            out(chalk_1.default.green(`  ${passed.length} passing`) + chalk_1.default.gray(` (${elapsedMs}ms)`));
        }
        if (failed.length > 0) {
            out(chalk_1.default.red(`  ${failed.length} failed`));
        }
        if (ignored.length > 0) {
            out(chalk_1.default.gray(`  ${ignored.length} ignored`));
        }
    }
    const stage = {
        ...testing,
        passed: passed.map(r => r.name),
        failed: failed.map(r => r.name),
        ignored: ignored.map(r => r.name),
        errors: [],
    };
    // Nothing failed, so we are OK, and can exit early
    if (failed.length === 0) {
        return (0, either_1.right)(stage);
    }
    // We know that there are errors, so report as required by the verbosity configuration
    if (verbosity_1.verbosity.hasTestDetails(verbosityLevel)) {
        const code = prev.sourceCode;
        const finders = (0, errorReporter_1.createFinders)(code);
        const columns = !prev.args.out ? (0, graphics_1.terminalWidth)() : 80;
        out('');
        namedErrors.forEach(([testResult, err], index) => {
            const details = (0, errorReporter_1.formatError)(code, finders, err);
            // output the header
            out(`  ${index + 1}) ${testResult.name}:`);
            const lines = details.split('\n');
            // output the first two lines in red
            lines.slice(0, 2).forEach(l => out(chalk_1.default.red('      ' + l)));
            if (verbosity_1.verbosity.hasActionTracking(verbosityLevel)) {
                out('');
                testResult.frames.forEach((f, index) => {
                    out(`[${chalk_1.default.bold('Frame ' + index)}]`);
                    const console = {
                        width: columns,
                        out: (s) => process.stdout.write(s),
                    };
                    (0, graphics_1.printExecutionFrameRec)(console, f, []);
                    out('');
                });
                if (testResult.frames.length == 0) {
                    out('    [No execution]');
                }
            }
            // output the seed
            out(chalk_1.default.gray(`    Use --seed=0x${testResult.seed.toString(16)} --match=${testResult.name} to repeat.`));
        });
        out('');
    }
    if (verbosity_1.verbosity.hasHints(options.verbosity) && !verbosity_1.verbosity.hasActionTracking(options.verbosity)) {
        out(chalk_1.default.gray(`\n  Use --verbosity=3 to show executions.`));
        out(chalk_1.default.gray(`  Further debug with: quint test --verbosity=3 ${prev.args.input}`));
    }
    return cliErr('Tests failed', stage);
}
exports.runTests = runTests;
// Print a counterexample if the appropriate verbosity is set
function maybePrintCounterExample(verbosityLevel, states, frames = []) {
    if (verbosity_1.verbosity.hasStateOutput(verbosityLevel)) {
        console.log(chalk_1.default.gray('An example execution:\n'));
        const myConsole = {
            width: (0, graphics_1.terminalWidth)(),
            out: (s) => process.stdout.write(s),
        };
        (0, graphics_1.printTrace)(myConsole, states, frames);
    }
}
function maybePrintWitnesses(verbosityLevel, evalResult, witnesses) {
    if (verbosity_1.verbosity.hasWitnessesOutput(verbosityLevel)) {
        evalResult.map(r => {
            if (r.witnessingTraces.length > 0) {
                console.log(chalk_1.default.green('Witnesses:'));
            }
            r.witnessingTraces.forEach((n, i) => {
                const percentage = chalk_1.default.gray(`(${(((1.0 * n) / r.samples) * 100).toFixed(2)}%)`);
                console.log(`${chalk_1.default.yellow(witnesses[i])} was witnessed in ${chalk_1.default.green(n)} trace(s) out of ${r.samples} explored ${percentage}`);
            });
        });
    }
}
/**
 * Run the simulator.
 *
 * @param prev the procedure stage produced by `typecheck`
 */
async function runSimulator(prev) {
    const simulator = { ...prev, stage: 'running' };
    const startMs = Date.now();
    // Force disable output if `--out-itf` is set
    const verbosityLevel = prev.args.outItf ? 0 : deriveVerbosity(prev.args);
    const mainName = guessMainModule(prev);
    const main = prev.modules.find(m => m.name === mainName);
    if (!main) {
        const error = { code: 'QNT405', message: `Main module ${mainName} not found` };
        return cliErr('Argument error', { ...prev, errors: [mkErrorMessage(prev.sourceMap)(error)] });
    }
    const rngOrError = mkRng(prev.args.seed);
    if (rngOrError.isLeft()) {
        return cliErr(rngOrError.value, { ...simulator, errors: [] });
    }
    const rng = rngOrError.unwrap();
    const options = {
        init: prev.args.init,
        step: prev.args.step,
        invariant: prev.args.invariant,
        maxSamples: prev.args.maxSamples,
        maxSteps: prev.args.maxSteps,
        rng,
        verbosity: verbosityLevel,
        storeMetadata: prev.args.mbt,
        numberOfTraces: prev.args.nTraces,
        onTrace: (index, status, vars, states) => {
            const itfFile = prev.args.outItf;
            if (itfFile) {
                const filename = expandOutputTemplate(itfFile, index, { autoAppend: prev.args.nTraces > 1 });
                const trace = (0, itf_1.toItf)(vars, states, prev.args.mbt);
                if (trace.isRight()) {
                    const jsonObj = addItfHeader(prev.args.input, status, trace.value);
                    writeToJson(filename, jsonObj);
                }
                else {
                    console.error(`ITF conversion failed on ${index}: ${trace.value}`);
                }
            }
        },
    };
    const recorder = (0, trace_1.newTraceRecorder)(options.verbosity, options.rng, options.numberOfTraces);
    function toExpr(input) {
        const parseResult = (0, quintParserFrontend_1.parseExpressionOrDeclaration)(input, '<input>', prev.idGen, prev.sourceMap);
        if (parseResult.kind !== 'expr') {
            return (0, either_1.left)({ code: 'QNT501', message: `Expected ${input} to be a valid expression` });
        }
        prev.resolver.switchToModule(mainName);
        (0, IRVisitor_1.walkExpression)(prev.resolver, parseResult.expr);
        if (prev.resolver.errors.length > 0) {
            return (0, either_1.left)(prev.resolver.errors[0]);
        }
        return (0, either_1.right)(parseResult.expr);
    }
    const argsParsingResult = (0, either_1.mergeInMany)([prev.args.init, prev.args.step, prev.args.invariant, ...prev.args.witnesses].map(toExpr));
    if (argsParsingResult.isLeft()) {
        return cliErr('Argument error', {
            ...simulator,
            errors: argsParsingResult.value.map(mkErrorMessage(new Map())),
        });
    }
    const [init, step, invariant, ...witnesses] = argsParsingResult.value;
    const evaluator = new evaluator_1.Evaluator(prev.resolver.table, recorder, options.rng, options.storeMetadata);
    const evalResult = evaluator.simulate(init, step, invariant, witnesses, prev.args.maxSamples, prev.args.maxSteps, prev.args.nTraces ?? 1);
    const elapsedMs = Date.now() - startMs;
    const outcome = evalResult.isRight()
        ? { status: evalResult.value.result.value ? 'ok' : 'violation' }
        : { status: 'error', errors: [evalResult.value] };
    const states = recorder.bestTraces[0]?.frame?.args?.map(e => e.toQuintEx(idGenerator_1.zerog));
    const frames = recorder.bestTraces[0]?.frame?.subframes;
    simulator.seed = recorder.bestTraces[0]?.seed;
    recorder.bestTraces.forEach((trace, index) => {
        const maybeEvalResult = trace.frame.result;
        if (maybeEvalResult.isLeft()) {
            return cliErr('Runtime error', {
                ...simulator,
                errors: [mkErrorMessage(simulator.sourceMap)(maybeEvalResult.value)],
            });
        }
        const quintExResult = maybeEvalResult.value.toQuintEx(prev.idGen);
        (0, assert_1.default)(quintExResult.kind === 'bool', 'invalid simulation produced non-boolean value ');
        const simulationSucceeded = quintExResult.value;
        const status = simulationSucceeded ? 'ok' : 'violation';
        const states = trace.frame.args.map(e => e.toQuintEx(prev.idGen));
        options.onTrace(index, status, evaluator.varNames(), states);
    });
    switch (outcome.status) {
        case 'error':
            return cliErr('Runtime error', {
                ...simulator,
                status: outcome.status,
                trace: states,
                errors: outcome.errors.map(mkErrorMessage(prev.sourceMap)),
            });
        case 'ok':
            maybePrintCounterExample(verbosityLevel, states, frames);
            if (verbosity_1.verbosity.hasResults(verbosityLevel)) {
                console.log(chalk_1.default.green('[ok]') + ' No violation found ' + chalk_1.default.gray(`(${elapsedMs}ms).`));
                if (verbosity_1.verbosity.hasHints(verbosityLevel)) {
                    console.log(chalk_1.default.gray('You may increase --max-samples and --max-steps.'));
                    console.log(chalk_1.default.gray('Use --verbosity to produce more (or less) output.'));
                }
            }
            maybePrintWitnesses(verbosityLevel, evalResult, prev.args.witnesses);
            return (0, either_1.right)({
                ...simulator,
                status: outcome.status,
                trace: states,
            });
        case 'violation':
            maybePrintCounterExample(verbosityLevel, states, frames);
            if (verbosity_1.verbosity.hasResults(verbosityLevel)) {
                console.log(chalk_1.default.red(`[violation]`) + ' Found an issue ' + chalk_1.default.gray(`(${elapsedMs}ms).`));
                if (verbosity_1.verbosity.hasHints(verbosityLevel)) {
                    console.log(chalk_1.default.gray('Use --verbosity=3 to show executions.'));
                }
            }
            maybePrintWitnesses(verbosityLevel, evalResult, prev.args.witnesses);
            return cliErr('Invariant violated', {
                ...simulator,
                status: outcome.status,
                trace: states,
                errors: [],
            });
    }
}
exports.runSimulator = runSimulator;
/**  Compile to a flattened module, that includes the special q::* declarations
 *
 * @param typechecked the output of a preceding type checking stage
 */
async function compile(typechecked) {
    const args = typechecked.args;
    const mainName = guessMainModule(typechecked);
    const main = typechecked.modules.find(m => m.name === mainName);
    if (!main) {
        return cliErr(`module ${mainName} does not exist`, { ...typechecked, errors: [], sourceCode: new Map() });
    }
    // Wrap init, step, invariant and temporal properties in other definitions,
    // to make sure they are not considered unused in the main module and,
    // therefore, ignored by the flattener
    const extraDefsAsText = [`action q::init = ${args.init}`, `action q::step = ${args.step}`];
    if (args.invariant) {
        extraDefsAsText.push(`val q::inv = and(${args.invariant})`);
    }
    if (args.temporal) {
        extraDefsAsText.push(`temporal q::temporalProps = and(${args.temporal})`);
    }
    const extraDefs = extraDefsAsText.map(d => (0, quintParserFrontend_1.parseDefOrThrow)(d, typechecked.idGen, new Map()));
    main.declarations.push(...extraDefs);
    // We have to update the lookup table and analysis result with the new definitions. This is not ideal, and the problem
    // is that is hard to add this definitions in the proper stage, in our current setup. We should try to tackle this
    // while solving #1052.
    const resolutionResult = (0, quintParserFrontend_1.parsePhase3importAndNameResolution)({ ...typechecked, errors: [] });
    if (resolutionResult.errors.length > 0) {
        const errors = resolutionResult.errors.map(mkErrorMessage(typechecked.sourceMap));
        return cliErr('name resolution failed', { ...typechecked, errors });
    }
    typechecked.table = resolutionResult.table;
    (0, quintAnalyzer_1.analyzeInc)(typechecked, typechecked.table, extraDefs);
    // Flatten modules, replacing instances, imports and exports with their definitions
    const { flattenedModules, flattenedTable, flattenedAnalysis } = (0, fullFlattener_1.flattenModules)(typechecked.modules, typechecked.table, typechecked.idGen, typechecked.sourceMap, typechecked);
    // Pick the main module
    const flatMain = flattenedModules.find(m => m.name === mainName);
    return (0, either_1.right)({
        ...typechecked,
        ...flattenedAnalysis,
        mainModule: flatMain,
        table: flattenedTable,
        stage: 'compiling',
    });
}
exports.compile = compile;
/** output a compiled spec in the format specified in the `compiled.args.target` to stdout
 *
 * @param compiled The result of a preceding compile stage
 */
async function outputCompilationTarget(compiled) {
    const stage = 'outputting target';
    const args = compiled.args;
    const verbosityLevel = deriveVerbosity(args);
    const parsedSpecJson = jsonStringOfOutputStage(pickOutputStage({ ...compiled, modules: [compiled.mainModule], table: compiled.table }));
    switch (compiled.args.target.toLowerCase()) {
        case 'json':
            process.stdout.write(parsedSpecJson);
            return (0, either_1.right)(compiled);
        case 'tlaplus': {
            const toTlaResult = await (0, compileToTlaplus_1.compileToTlaplus)(args.serverEndpoint, args.apalacheVersion, parsedSpecJson, verbosityLevel);
            return toTlaResult
                .mapRight(tla => {
                process.stdout.write(tla); // Write out, since all went right
                return compiled;
            })
                .mapLeft(err => {
                return {
                    msg: err.explanation,
                    stage: { ...compiled, stage, status: 'error', errors: err.errors },
                };
            });
        }
        default:
            // This is validated in the arg parsing
            (0, assert_1.fail)(`Invalid option for --target`);
    }
}
exports.outputCompilationTarget = outputCompilationTarget;
/**
 * Verify a spec via Apalache.
 *
 * @param prev the procedure stage produced by `typecheck`
 */
async function verifySpec(prev) {
    const verifying = { ...prev, stage: 'verifying' };
    const args = verifying.args;
    // Force disable output if `--out-itf` is set
    const verbosityLevel = prev.args.outItf ? 0 : deriveVerbosity(prev.args);
    const itfFile = prev.args.outItf;
    if (itfFile) {
        if (itfFile.includes(PLACEHOLDERS.test) || itfFile.includes(PLACEHOLDERS.seq)) {
            console.log(`${chalk_1.default.yellow('[warning]')} the output file contains ${chalk_1.default.grey(PLACEHOLDERS.test)} or ${chalk_1.default.grey(PLACEHOLDERS.seq)}, but this has no effect since at most a single trace will be produced.`);
        }
    }
    let loadedConfig = {};
    try {
        if (args.apalacheConfig) {
            loadedConfig = JSON.parse((0, fs_1.readFileSync)(args.apalacheConfig, 'utf-8'));
        }
    }
    catch (err) {
        return cliErr(`failed to read Apalache config: ${err.message}`, { ...verifying, errors: [], sourceCode: new Map() });
    }
    const veryfiyingFlat = { ...prev, modules: [prev.mainModule] };
    const parsedSpec = jsonStringOfOutputStage(pickOutputStage(veryfiyingFlat));
    // We need to insert the data form CLI args into their appropriate locations
    // in the Apalache config
    const config = {
        ...loadedConfig,
        input: {
            ...(loadedConfig.input ?? {}),
            source: {
                type: 'string',
                format: 'qnt',
                content: parsedSpec,
            },
        },
        checker: {
            ...(loadedConfig.checker ?? {}),
            length: args.maxSteps,
            init: 'q::init',
            next: 'q::step',
            inv: args.invariant ? ['q::inv'] : undefined,
            'temporal-props': args.temporal ? ['q::temporalProps'] : undefined,
            tuning: {
                ...(loadedConfig.checker?.tuning ?? {}),
                'search.simulation': args.randomTransitions ? 'true' : 'false',
            },
        },
    };
    const startMs = Date.now();
    return (0, quintVerifier_1.verify)(args.serverEndpoint, args.apalacheVersion, config, verbosityLevel).then(res => {
        const elapsedMs = Date.now() - startMs;
        return res
            .map(_ => {
            if (verbosity_1.verbosity.hasResults(verbosityLevel)) {
                console.log(chalk_1.default.green('[ok]') + ' No violation found ' + chalk_1.default.gray(`(${elapsedMs}ms).`));
                if (verbosity_1.verbosity.hasHints(verbosityLevel)) {
                    console.log(chalk_1.default.gray('You may increase --max-steps.'));
                    console.log(chalk_1.default.gray('Use --verbosity to produce more (or less) output.'));
                }
            }
            return { ...verifying, status: 'ok', errors: [] };
        })
            .mapLeft(err => {
            const trace = err.traces ? (0, itf_1.ofItf)(err.traces[0]) : undefined;
            const status = trace !== undefined ? 'violation' : 'failure';
            if (trace !== undefined) {
                // Always print the conterexample, unless the output is being directed to one of the outfiles
                maybePrintCounterExample(verbosityLevel, trace);
                if (verbosity_1.verbosity.hasResults(verbosityLevel)) {
                    console.log(chalk_1.default.red(`[${status}]`) + ' Found an issue ' + chalk_1.default.gray(`(${elapsedMs}ms).`));
                }
                if (prev.args.outItf && err.traces) {
                    writeToJson(prev.args.outItf, err.traces[0]);
                }
            }
            return {
                msg: err.explanation,
                stage: { ...verifying, status, errors: err.errors, trace },
            };
        });
    });
}
exports.verifySpec = verifySpec;
/**
 * Produces documentation from docstrings in a Quint specification.
 *
 * @param loaded the procedure stage produced by `load`
 */
async function docs(loaded) {
    const { sourceCode, path } = loaded;
    const text = sourceCode.get(path);
    const parsing = { ...loaded, stage: 'documentation' };
    const phase1Data = (0, quintParserFrontend_1.parsePhase1fromText)((0, idGenerator_1.newIdGenerator)(), text, path);
    const allEntries = phase1Data.modules.map(module => {
        const documentationEntries = (0, docs_1.produceDocs)(module);
        const title = `# Documentation for ${module.name}\n\n`;
        const markdown = title + [...documentationEntries.values()].map(docs_1.toMarkdown).join('\n\n');
        console.log(markdown);
        return [module.name, documentationEntries];
    });
    if (phase1Data.errors.length > 0) {
        const newErrorMessages = phase1Data.errors.map(mkErrorMessage(phase1Data.sourceMap));
        const errorMessages = parsing.errors ? parsing.errors.concat(newErrorMessages) : newErrorMessages;
        return (0, either_1.left)({ msg: 'parsing failed', stage: { ...parsing, errors: errorMessages } });
    }
    return (0, either_1.right)({ ...parsing, documentation: new Map(allEntries) });
}
exports.docs = docs;
/** Write the OutputStage of the procedureStage as JSON, if --out is set
 * Otherwise, report any stage errors to STDOUT
 */
function outputResult(result) {
    result
        .map(stage => {
        const verbosityLevel = deriveVerbosity(stage.args);
        const outputData = pickOutputStage(stage);
        if (stage.args.out) {
            writeToJson(stage.args.out, outputData);
        }
        else if (!stage.args.outItf && outputData.seed && verbosity_1.verbosity.hasResults(verbosityLevel)) {
            console.log(chalk_1.default.gray(`Use --seed=0x${outputData.seed.toString(16)} to reproduce.`));
        }
        process.exit(0);
    })
        .mapLeft(({ msg, stage }) => {
        const { args, errors, sourceCode } = stage;
        const verbosityLevel = deriveVerbosity(args);
        const outputData = pickOutputStage(stage);
        if (args.out) {
            writeToJson(args.out, outputData);
        }
        else {
            const finders = (0, errorReporter_1.createFinders)(sourceCode);
            (0, lodash_1.uniqWith)(errors, lodash_1.isEqual).forEach(err => console.error((0, errorReporter_1.formatError)(sourceCode, finders, err)));
            if (!stage.args.outItf && outputData.seed && verbosity_1.verbosity.hasResults(verbosityLevel)) {
                console.log(chalk_1.default.gray(`Use --seed=0x${outputData.seed.toString(16)} to reproduce.`));
            }
            console.error(`error: ${msg}`);
        }
        process.exit(1);
    });
}
exports.outputResult = outputResult;
function guessMainModule(stage) {
    if (stage.args.main) {
        // the main module is specified via --main
        return stage.args.main;
    }
    if (stage.defaultModuleName.isJust()) {
        // there is only one module in the source file, make it main
        return stage.defaultModuleName.unwrap();
    }
    // guess the name from the filename
    return (0, path_1.basename)(stage.args.input, '.qnt');
}
/**
 * Produce a random-number generator: Either a predictable one using a seed,
 * or a reasonably unpredictable one.
 */
function mkRng(seedText) {
    let seed;
    if (seedText !== undefined) {
        // since yargs does not has a type for big integers,
        // we do it with a fallback
        try {
            seed = BigInt(seedText);
        }
        catch (SyntaxError) {
            return (0, either_1.left)(`--seed must be a big integer, found: ${seedText}`);
        }
    }
    return (0, either_1.right)(seed ? (0, rng_1.newRng)(seed) : (0, rng_1.newRng)());
}
function addItfHeader(source, status, traceInJson) {
    return {
        '#meta': {
            format: 'ITF',
            'format-description': 'https://apalache-mc.org/docs/adr/015adr-trace.html',
            source,
            status,
            description: 'Created by Quint on ' + new Date(),
            timestamp: Date.now(),
        },
        ...traceInJson,
    };
}
// Preprocess troublesome types so they are represented in JSON.
//
// We need it particularly because, by default, serialization of Map and Set
// objects just produces an empty object
// (see https://stackoverflow.com/questions/46634449/json-stringify-of-object-of-map-return-empty)
//
// The approach here follows https://stackoverflow.com/a/56150320/1187277
function replacer(_key, value) {
    if (value instanceof Map) {
        // Represent Maps as JSON objects
        return Object.fromEntries(value);
    }
    else if (value instanceof Set) {
        // Represent Sets as JSON arrays
        return Array.from(value);
    }
    else {
        return value;
    }
}
function jsonStringOfOutputStage(json) {
    return json_bigint_1.default.stringify(json, replacer);
}
/**
 * Write json to a file.
 *
 * @param filename name of the file to write to
 * @param json is an object tree to write
 */
function writeToJson(filename, json) {
    const path = (0, path_1.resolve)((0, process_1.cwd)(), filename);
    (0, fs_1.writeFileSync)(path, jsonStringOfOutputStage(json));
}
/**
 * Does a definition name match the expected test criteria.
 *
 * @param tests an optional array of test names
 * @param name the name of a definition to match
 * @returns whether the name matches the tests, if tests are not undefined,
 *          or name ends with 'Test'
 *
 */
function isMatchingTest(match, name) {
    if (match) {
        return new RegExp(match).exec(name) !== null;
    }
    else {
        return name.endsWith('Test');
    }
}
// Derive the verbosity for simulation and verification routines
function deriveVerbosity(args) {
    return args.out ? 0 : args.verbosity;
}
const PLACEHOLDERS = {
    test: '{test}',
    seq: '{seq}',
};
/**
 * Expand the output template with the name of the test and the index of the trace.
 *
 * Possible placeholders:
 * - {test} is replaced with the name of the test
 * - {seq} is replaced with the index of the trace
 *
 * If {seq} is not present and `options.autoAppend` is true,
 * the index is appended to the filename, before the extension.
 *
 * @param template the output template
 * @param name the name of the test
 * @param index the index of the trace
 * @param options An object of the form `{ autoAppend: boolean }`
 * @returns the expanded output template
 */
function expandNamedOutputTemplate(template, name, index, options) {
    return expandOutputTemplate(template.replaceAll(PLACEHOLDERS.test, name), index, options);
}
/**
 * Expand the output template with the index of the trace.
 *
 * The {seq} placeholder is replaced with the index of the trace.
 *
 * If {seq} is not present and `options.autoAppend` is true,
 * the index is appended to the filename, before the extension.
 *
 * @param template the output template
 * @param index the index of the trace
 * @param options An object of the form `{ autoAppend: boolean }`
 * @returns the expanded output template
 */
function expandOutputTemplate(template, index, options) {
    if (template.includes(PLACEHOLDERS.seq)) {
        return template.replaceAll(PLACEHOLDERS.seq, index.toString());
    }
    if (options.autoAppend) {
        const parts = template.split('.');
        parts[0] += `${index}`;
        return parts.join('.');
    }
    return template;
}
//# sourceMappingURL=cliCommands.js.map