import * as p from '../generated/QuintParser';
import { IdGenerator } from '../idGenerator';
import { QuintListener } from '../generated/QuintListener';
import { QuintDeclaration, QuintEx, QuintLambdaParameter, QuintModule } from '../ir/quintIr';
import { QuintType, Row, RowField } from '../ir/quintTypes';
import { SourceMap } from './quintParserFrontend';
import { QuintError } from '../quintError';
/**
 * An ANTLR4 listener that constructs QuintIr objects out of the abstract
 * syntax tree. This listener does the minimal work of translating the AST
 * into IR. All semantic checks and type checking must be done at later
 * phases, as the IR may be constructed by other means.
 *
 * The listener should be able to execute on top of malformed ASTs. It produces
 * undefined/default components to fill in the gaps. The issues found while
 * producing an AST should be reported, but should not prevent an IR to be
 * constructed. This allow us to gather additional static analysis information
 * to report before aborting.
 *
 * @author Igor Konnov, Gabriela Moreira
 */
export declare class ToIrListener implements QuintListener {
    constructor(sourceLocation: string, idGen: IdGenerator);
    /**
     * The modules derived as a result of parsing
     */
    modules: QuintModule[];
    /**
     * The stack of types that can be used as a result of type parsing.
     */
    typeStack: QuintType[];
    sourceMap: SourceMap;
    /**
     * If errors occur in the listener, this array contains explanations.
     */
    errors: QuintError[];
    private sourceLocation;
    protected declarationStack: QuintDeclaration[];
    protected exprStack: QuintEx[];
    protected paramStack: QuintLambdaParameter[];
    protected identOrStarStack: string[];
    protected rowStack: Row[];
    protected variantStack: RowField[];
    protected idGen: IdGenerator;
    exitModule(ctx: p.ModuleContext): void;
    exitConst(ctx: p.ConstContext): void;
    exitVar(ctx: p.VarContext): void;
    exitLetIn(ctx: p.LetInContext): void;
    /** **************** translate operator definititons **********************/
    exitAnnotatedOperDef(ctx: p.AnnotatedOperDefContext): void;
    exitDeprecatedOperDef(ctx: p.DeprecatedOperDefContext): void;
    private putDeclarationOnStack;
    exitAssume(ctx: p.AssumeContext): void;
    exitImportMod(ctx: p.ImportModContext): void;
    exitExportMod(ctx: p.ExportModContext): void;
    exitTypeAbstractDef(ctx: p.TypeAbstractDefContext): void;
    exitTypeAliasDef(ctx: p.TypeAliasDefContext): void;
    exitTypeSumDef(ctx: p.TypeSumDefContext): void;
    exitTypeSumVariant(ctx: p.TypeSumVariantContext): void;
    private checkForUndeclaredTypeVariables;
    exitInstanceMod(ctx: p.InstanceModContext): void;
    /** ******************* translate expressions **************************/
    exitLiteralOrId(ctx: p.LiteralOrIdContext): void;
    exitListApp(ctx: any): void;
    exitOperApp(ctx: p.OperAppContext): void;
    exitDotCall(ctx: p.DotCallContext): void;
    exitArgList(ctx: p.ArgListContext): void;
    exitLambdaUnsugared(ctx: p.LambdaUnsugaredContext): void;
    exitLambdaTupleSugar(ctx: p.LambdaTupleSugarContext): void;
    exitParameter(ctx: p.ParameterContext): void;
    exitAnnotatedParameter(ctx: p.AnnotatedParameterContext): void;
    exitIdentOrStar(ctx: p.IdentOrStarContext): void;
    exitTuple(ctx: p.TupleContext): void;
    exitUnit(ctx: p.UnitContext): void;
    exitPair(ctx: p.PairContext): void;
    exitList(ctx: p.ListContext): void;
    exitRecElem(ctx: p.RecElemContext): void;
    exitRecord(ctx: p.RecordContext): void;
    exitPlusMinus(ctx: p.PlusMinusContext): void;
    exitMultDiv(ctx: p.MultDivContext): void;
    exitPow(ctx: any): void;
    exitUminus(ctx: any): void;
    exitAsgn(ctx: p.AsgnContext): void;
    exitRelations(ctx: p.RelationsContext): void;
    exitErrorEq(ctx: p.ErrorEqContext): void;
    exitAnd(ctx: any): void;
    exitOr(ctx: any): void;
    exitImplies(ctx: any): void;
    exitIff(ctx: any): void;
    exitAndExpr(ctx: p.AndExprContext): void;
    exitOrExpr(ctx: p.OrExprContext): void;
    exitActionAll(ctx: p.ActionAllContext): void;
    exitActionAny(ctx: p.ActionAnyContext): void;
    exitIfElse(ctx: p.IfElseContext): void;
    exitMatchSumExpr(ctx: p.MatchSumExprContext): void;
    private formMatchCase;
    /** ******************* translate types ********************************/
    exitTypeInt(ctx: p.TypeIntContext): void;
    exitTypeBool(ctx: p.TypeBoolContext): void;
    exitTypeStr(ctx: p.TypeStrContext): void;
    exitTypeVar(ctx: p.TypeVarContext): void;
    exitTypeConst(ctx: p.TypeConstContext): void;
    exitTypeApp(ctx: p.TypeAppContext): void;
    exitTypeSet(ctx: p.TypeSetContext): void;
    exitTypeList(ctx: p.TypeListContext): void;
    exitTypeFun(ctx: p.TypeFunContext): void;
    exitTypeUnit(ctx: p.TypeUnitContext): void;
    exitTypeTuple(ctx: p.TypeTupleContext): void;
    exitRow(ctx: p.RowContext): void;
    exitTypeRec(ctx: p.TypeRecContext): void;
    exitTypeOper(ctx: p.TypeOperContext): void;
    exitDocumentedDeclaration(ctx: p.DocumentedDeclarationContext): void;
    private getId;
    private loc;
    private pushApplication;
    private popType;
    private popRow;
    private undefinedExpr;
    private undefinedDeclaration;
    private undefinedParam;
    private undefinedType;
    private undefinedVariant;
    /**
     * Return a function that takes a QuintEx `body`, a QuintLambdaParameter `param`,
     * and the parameter's `index` and returns a let-in definition of the form
     * `let ${param.name} = ${freshParam.name}._${index} ; body`.
     *
     * Intended use: Reduce tupled Quint lambda parameters into an equivalent let-in
     * form ("tuple unboxing"). For example, `((a, b)) => body` can be reduced to
     * `freshParam => let a = freshParam._1; let b = freshParam._2; body`.
     *
     * @param ctx         parser context of the tuple-sugared lambda
     * @param freshParam  a fresh QuintLambdaParameter to substitute for the tupled parameters
     */
    private letBindingForTupleParam;
    private checkForUppercaseTypeName;
}
