"use strict";
/* ----------------------------------------------------------------------------------
 * Copyright 2022-2024 Informal Systems
 * Licensed under the Apache License, Version 2.0.
 * See LICENSE in the project root for license information.
 * --------------------------------------------------------------------------------- */
Object.defineProperty(exports, "__esModule", { value: true });
exports.ConstraintGeneratorVisitor = void 0;
const quintIr_1 = require("../ir/quintIr");
const quintTypes_1 = require("../ir/quintTypes");
const IRprinting_1 = require("../ir/IRprinting");
const either_1 = require("@sweet-monads/either");
const errorTree_1 = require("../errorTree");
const builtinSignatures_1 = require("./builtinSignatures");
const base_1 = require("./base");
const substitutions_1 = require("./substitutions");
const specialConstraints_1 = require("./specialConstraints");
const FreshVarGenerator_1 = require("../FreshVarGenerator");
// `validateArity(opName, args, pred, msg)` is `right(null)` if
// if `pred(args.length) === true`, and otherwise `left(err)`, where `err`
// is constructed using the given `opName` and `msg`.
//
// `msg` should contain a textual description of the expected argument
// length, e.g., ("1", "2", "even number of", ...).
//
// Use this for operators that cannot be typed in the Quint type system.
function validateArity(opcode, args, pred, msg) {
    if (!pred(args.length)) {
        return (0, either_1.left)((0, errorTree_1.buildErrorLeaf)(`Checking arity for application of ${opcode}`, `Operator expects ${msg} arguments but was given ${args.length}`));
    }
    else {
        return (0, either_1.right)(null);
    }
}
// A visitor that collects types and constraints for a module's expressions
//
// NOTE: Assumes all type applications have been resolved by typeApplicationResolution first
class ConstraintGeneratorVisitor {
    // Inject dependency to allow manipulation in unit tests
    constructor(solvingFunction, table, types) {
        this.types = new Map();
        this.errors = new Map();
        this.constraints = [];
        this.builtinSignatures = (0, builtinSignatures_1.getSignatures)();
        // A map to save which type variables were free when we started visiting an opdef or an assume
        this.tvs = new Map();
        // Temporary type map only for types in scope for a certain declaration
        this.typesInScope = new Map();
        // Track location descriptions for error tree traces
        this.location = '';
        this.solvingFunction = solvingFunction;
        this.table = table;
        this.freshVarGenerator = new FreshVarGenerator_1.FreshVarGenerator();
        if (types) {
            this.types = types;
        }
    }
    getResult() {
        return [this.errors, this.types];
    }
    enterExpr(e) {
        this.location = `Generating constraints for ${(0, IRprinting_1.expressionToString)(e)}`;
    }
    exitVar(e) {
        this.addToResults(e.id, (0, either_1.right)((0, base_1.toScheme)(e.typeAnnotation)));
    }
    exitConst(e) {
        this.addToResults(e.id, (0, either_1.right)((0, base_1.toScheme)(e.typeAnnotation)));
    }
    exitInstance(def) {
        if (this.errors.size !== 0) {
            return;
        }
        // For each override, ensure that the the type for the name and the type of
        // the value are the same
        def.overrides.forEach(([name, ex]) => {
            this.addToResults(name.id, this.typeForName(name.name, name.id, 2).map(base_1.toScheme));
            this.fetchResult(name.id).chain(originalType => {
                return this.fetchResult(ex.id).map(expressionType => {
                    this.constraints.push({ kind: 'eq', types: [originalType.type, expressionType.type], sourceId: ex.id });
                });
            });
        });
        // Solve constraints here since this won't go through `exitDef`
        if (this.constraints.length > 0) {
            this.solveConstraints();
        }
    }
    //     n: t ∈ Γ
    // ----------------- (NAME)
    //  Γ ⊢ n: (t, true)
    exitName(e) {
        if (this.errors.size !== 0) {
            return;
        }
        this.addToResults(e.id, this.typeForName(e.name, e.id, 2).map(base_1.toScheme));
    }
    // Literals have always the same type and the empty constraint
    exitLiteral(e) {
        this.addToResults(e.id, (0, either_1.right)((0, base_1.toScheme)({ kind: e.kind })));
    }
    //   op: q ∈ Γ   Γ ⊢  p0, ..., pn: (t0, c0), ..., (tn, cn)   a is fresh
    // ------------------------------------------------------------------------ (APP)
    //    Γ ⊢ op(p0, ..., pn): (a, q ~ (t0, ..., tn) => a ∧ c0 ∧ ... ∧ cn)
    exitApp(e) {
        if (this.errors.size !== 0) {
            return;
        }
        const argsResult = (0, either_1.mergeInMany)(e.args.map(e => {
            return this.fetchResult(e.id).map(r => [e, r.type]);
        }));
        // We want `definedSignature` computed before the fresh variable `a` so that the
        // numbering of their fresh variables stays in order, with `a`, used for return types,
        // bearing the highest number.
        const definedSignature = this.typeForName(e.opcode, e.id, e.args.length);
        const a = { kind: 'var', name: this.freshVarGenerator.freshVar('_t') };
        const result = argsResult
            .chain(results => {
            switch (e.opcode) {
                // Record operators
                case 'Rec':
                    return validateArity(e.opcode, results, l => l % 2 === 0, 'even number of').chain(() => (0, specialConstraints_1.recordConstructorConstraints)(e.id, results, a));
                case 'field':
                    return validateArity(e.opcode, results, l => l === 2, '2').chain(() => (0, specialConstraints_1.fieldConstraints)(e.id, results, a));
                case 'fieldNames':
                    return validateArity(e.opcode, results, l => l === 1, '1').chain(() => (0, specialConstraints_1.fieldNamesConstraints)(e.id, results, a));
                case 'with':
                    return validateArity(e.opcode, results, l => l === 3, '3').chain(() => (0, specialConstraints_1.withConstraints)(e.id, results, a));
                // Tuple operators
                case 'Tup':
                    return (0, specialConstraints_1.tupleConstructorConstraints)(e.id, results, a);
                case 'item':
                    return validateArity(e.opcode, results, l => l === 2, '2').chain(() => (0, specialConstraints_1.itemConstraints)(e.id, results, a));
                // Sum type operators
                case 'variant':
                    return validateArity(e.opcode, results, l => l === 2, '2').chain(() => (0, specialConstraints_1.variantConstraints)(e.id, results, a));
                case 'matchVariant':
                    return validateArity(e.opcode, results, l => l % 2 !== 0, 'odd number of').chain(() => (0, specialConstraints_1.matchConstraints)(e.id, results, a));
                // Otherwise it's a standard operator with a definition in the context
                default:
                    return definedSignature.map(t1 => {
                        const t2 = { kind: 'oper', args: results.map(r => r[1]), res: a };
                        const c = { kind: 'eq', types: [t1, t2], sourceId: e.id };
                        return [c];
                    });
            }
        })
            .map(cs => {
            this.constraints.push(...cs);
            return (0, base_1.toScheme)(a);
        });
        this.addToResults(e.id, result);
    }
    enterLambda(expr) {
        expr.params.forEach(p => {
            const varName = p.name === '_' ? this.freshVarGenerator.freshVar('_t') : `t_${p.name}_${p.id}`;
            const paramTypeVar = { kind: 'var', name: varName };
            this.addToResults(p.id, (0, either_1.right)((0, base_1.toScheme)(paramTypeVar)));
            if (p.typeAnnotation) {
                this.constraints.push({ kind: 'eq', types: [paramTypeVar, p.typeAnnotation], sourceId: p.id });
            }
        });
    }
    //    Γ ∪ {p0: t0, ..., pn: tn} ⊢ e: (te, c)    t0, ..., tn are fresh
    // ---------------------------------------------------------------------- (LAMBDA)
    //            Γ ⊢ (p0, ..., pn) => e: ((t0, ..., tn) => te, c)
    exitLambda(e) {
        if (this.errors.size !== 0) {
            return;
        }
        const result = this.fetchResult(e.expr.id).chain(resultType => {
            const paramTypes = (0, either_1.mergeInMany)(e.params.map(p => this.fetchResult(p.id).map(e => this.newInstance(e))));
            return paramTypes
                .map((ts) => {
                const newType = { kind: 'oper', args: ts, res: resultType.type };
                return (0, base_1.toScheme)(newType);
            })
                .mapLeft(e => {
                throw new Error(`This should be impossible: Lambda variables not found: ${e.map(errorTree_1.errorTreeToString)}`);
            });
        });
        this.addToResults(e.id, result);
    }
    //   Γ ⊢ e1: (t1, c1)  s = solve(c1)     s(Γ ∪ {n: t1}) ⊢ e2: (t2, c2)
    // ------------------------------------------------------------------------ (LET-OPDEF)
    //               Γ ⊢ val n = e1 { e2 }: (quantify(t2), c1 ∧ c2)
    exitLet(e) {
        if (this.errors.size !== 0) {
            return;
        }
        this.addToResults(e.id, this.fetchResult(e.expr.id));
    }
    exitDecl(_def) {
        this.typesInScope = new Map();
    }
    enterOpDef(def) {
        // Save which type variables were free when we started visiting this op def
        const tvs = this.freeNamesInScope();
        this.tvs.set(def.id, tvs);
    }
    exitOpDef(def) {
        if (this.errors.size !== 0) {
            return;
        }
        this.fetchResult(def.expr.id).map(t => {
            if (def.typeAnnotation) {
                this.constraints.push({ kind: 'eq', types: [t.type, def.typeAnnotation], sourceId: def.id });
            }
        });
        const tvs_before = this.tvs.get(def.id);
        if (this.constraints.length > 0) {
            this.solveConstraints().map(subs => {
                // For every free name we are binding in the substitutions, the names occuring in the value of the substitution
                // have to become free as well.
                addBindingsToFreeTypes(tvs_before, subs);
                if ((0, quintIr_1.isAnnotatedDef)(def)) {
                    checkAnnotationGenerality(subs, def.typeAnnotation).mapLeft(err => this.errors.set(def.typeAnnotation?.id ?? def.id, err));
                }
            });
        }
        const tvs = this.freeNamesInScope();
        // Any new free names, that were not free before, have to be quantified
        const toQuantify = variablesDifference(tvs, tvs_before);
        this.fetchResult(def.expr.id).map(t => {
            this.addToResults(def.id, (0, either_1.right)(quantify(toQuantify, t.type)));
        });
    }
    enterAssume(e) {
        const tvs = this.freeNamesInScope();
        this.tvs.set(e.id, tvs);
    }
    exitAssume(e) {
        if (this.errors.size !== 0) {
            return;
        }
        this.fetchResult(e.assumption.id).map(t => {
            const tvs_before = this.tvs.get(e.id);
            const tvs = this.freeNamesInScope();
            const toQuantify = variablesDifference(tvs, tvs_before);
            this.addToResults(e.id, (0, either_1.right)(quantify(toQuantify, t.type)));
            this.constraints.push({ kind: 'eq', types: [t.type, { kind: 'bool' }], sourceId: e.id });
        });
    }
    addToResults(exprId, result) {
        result
            .mapLeft(err => this.errors.set(exprId, (0, errorTree_1.buildErrorTree)(this.location, err)))
            .map(r => {
            this.typesInScope.set(exprId, r);
            this.types.set(exprId, r);
        });
    }
    fetchResult(id) {
        const successfulResult = this.types.get(id);
        const failedResult = this.errors.get(id);
        if (failedResult) {
            return (0, either_1.left)(failedResult);
        }
        else if (successfulResult) {
            return (0, either_1.right)(successfulResult);
        }
        else {
            throw new Error(`Couldn't find any result for id ${id} while ${this.location}`);
        }
    }
    solveConstraints() {
        const constraint = { kind: 'conjunction', constraints: this.constraints, sourceId: 0n };
        // Remove solved constraints
        this.constraints = [];
        return this.solvingFunction(this.table, constraint)
            .mapLeft(errors => errors.forEach((err, id) => this.errors.set(id, err)))
            .map(subs => {
            this.typesInScope.forEach((oldScheme, id) => {
                const newType = (0, substitutions_1.applySubstitution)(this.table, subs, oldScheme.type);
                const newScheme = { ...oldScheme, type: newType };
                this.addToResults(id, (0, either_1.right)(newScheme));
            });
            return subs;
        });
    }
    typeForName(name, nameId, arity) {
        // Assumes a valid number of arguments
        if (this.builtinSignatures.has(name)) {
            const signatureFunction = this.builtinSignatures.get(name);
            const signature = signatureFunction(arity);
            return (0, either_1.right)(this.newInstance(signature));
        }
        else {
            const def = this.table.get(nameId);
            // FIXME: We have to check if the annotation is too general for var and consts as well
            // https://github.com/informalsystems/quint/issues/691
            if (def?.typeAnnotation) {
                return (0, either_1.right)(def.typeAnnotation);
            }
            const id = def?.id;
            if (!def || !id) {
                return (0, either_1.left)((0, errorTree_1.buildErrorLeaf)(this.location, `Signature not found for name: ${name}`));
            }
            return this.fetchResult(id).map(t => this.newInstance(t));
        }
    }
    newInstance(t) {
        const typeNames = Array.from(t.typeVariables);
        const rowNames = Array.from(t.rowVariables);
        const typeSubs = typeNames.map(name => {
            return { kind: 'type', name: name, value: { kind: 'var', name: this.freshVarGenerator.freshVar('_t') } };
        });
        const rowSubs = rowNames.map(name => {
            return { kind: 'row', name: name, value: { kind: 'var', name: this.freshVarGenerator.freshVar('_t') } };
        });
        const subs = (0, substitutions_1.compose)(this.table, typeSubs, rowSubs);
        return (0, substitutions_1.applySubstitution)(this.table, subs, t.type);
    }
    freeNamesInScope() {
        return [...this.typesInScope.values()].reduce((acc, t) => {
            const names = freeTypes(t);
            return {
                typeVariables: new Set([...names.typeVariables, ...acc.typeVariables]),
                rowVariables: new Set([...names.rowVariables, ...acc.rowVariables]),
            };
        }, { typeVariables: new Set(), rowVariables: new Set() });
    }
}
exports.ConstraintGeneratorVisitor = ConstraintGeneratorVisitor;
function checkAnnotationGenerality(subs, typeAnnotation) {
    if (!typeAnnotation) {
        return (0, either_1.right)(subs);
    }
    // Find type and row variables in the annotation that are bound to non-variable types in the substitutions.
    // This indicates that they are inferred to have a concrete type.
    const names = (0, quintTypes_1.typeNames)(typeAnnotation);
    const tooGeneralTypes = subs.filter(s => s.kind === 'type' && s.value.kind !== 'var' && names.typeVariables.has(s.name));
    const tooGeneralRows = subs.filter(s => s.kind === 'row' && s.value.kind !== 'var' && names.rowVariables.has(s.name));
    const errors = [...tooGeneralTypes, ...tooGeneralRows].map(s => {
        const expected = s.kind === 'type' ? (0, IRprinting_1.typeToString)(s.value) : (0, IRprinting_1.rowToString)(s.value);
        return (0, errorTree_1.buildErrorLeaf)(`Checking variable ${s.name}`, `Type annotation is too general: ${s.name} should be ${expected}`);
    });
    if (errors.length > 0) {
        return (0, either_1.left)((0, errorTree_1.buildErrorTree)(`Checking type annotation ${(0, IRprinting_1.typeToString)(typeAnnotation)}`, errors));
    }
    else {
        return (0, either_1.right)(subs);
    }
}
function quantify(tvs, type) {
    return { ...tvs, type };
}
function freeTypes(t) {
    const allNames = (0, quintTypes_1.typeNames)(t.type);
    return variablesDifference(allNames, { typeVariables: t.typeVariables, rowVariables: t.rowVariables });
}
function addBindingsToFreeTypes(free, substitutions) {
    // Assumes substitutions are topologically sorted, i.e. [ t0 |-> (t1, t2), t1 |-> (t3, t4) ]
    substitutions.forEach(s => {
        switch (s.kind) {
            case 'type':
                if (free.typeVariables.has(s.name)) {
                    const names = (0, quintTypes_1.typeNames)(s.value);
                    names.typeVariables.forEach(v => free.typeVariables.add(v));
                    names.rowVariables.forEach(v => free.rowVariables.add(v));
                }
                return;
            case 'row':
                if (free.rowVariables.has(s.name)) {
                    (0, quintTypes_1.rowNames)(s.value).forEach(v => free.rowVariables.add(v));
                }
                return;
        }
    });
}
function variablesDifference(a, b) {
    return {
        typeVariables: new Set([...a.typeVariables].filter(tv => !b.typeVariables.has(tv))),
        rowVariables: new Set([...a.rowVariables].filter(tv => !b.rowVariables.has(tv))),
    };
}
//# sourceMappingURL=constraintGenerator.js.map